from pathlib import PurePosixPath, Path
from subprocess import run

import auto_dlp.adb as adb
import auto_dlp.file_locations as fs


def _android_path_to_os(config, path: PurePosixPath):
    path = path.relative_to(config.adb_push_dir)
    return fs.music_dir() / Path(path)


def _push_allowed_files_list(config, artist: str, path: PurePosixPath, files: list):
    temp_file: Path = fs.adb_push_files()
    fs.touch_file(temp_file)

    with open(temp_file, "w") as file:
        for f in files:
            print(config.adb_push_dir / artist / PurePosixPath(f), file=file)

    run([adb.ADB_EXECUTABLE(), "-d", "push", str(temp_file.resolve()), str(path)])


def _allowed_files(config, artist: str):
    root = fs.artist_dir(artist)
    files_list = []

    def recurse(path: Path):
        files_list.append(path.relative_to(root))
        if path.is_dir():
            for sub in path.iterdir():
                recurse(sub)

    recurse(root)

    return files_list


def clean(config, artist: str, file: PurePosixPath):
    allowed_files = config.adb_push_dir / "allowed files.txt"
    sorted_allowed_files = config.adb_push_dir / "sorted allowed files.txt"
    _push_allowed_files_list(config, artist, allowed_files, _allowed_files(config, artist))
    run([adb.ADB_EXECUTABLE(), "shell",
         f'''
sort "{allowed_files}" > "{sorted_allowed_files}"

allowed_files=()

while read -r file; do
    allowed_files+=("$file");
done < "{sorted_allowed_files}"

length=${"{#allowed_files[@]}"}

find "{config.adb_push_dir}/{artist}" | while read -r file; do
is_allowed=false;
i=0;
while [[ $i -le $length ]] ; do
    compare_to="${"{allowed_files[$i]}"}"
    if [[ "$file" == "$compare_to" ]]; then
        # echo Yes $file $compare_to;
        is_allowed=true;
    fi
    ((i++))
done
if ! $is_allowed; then
    echo Deleting $file;
    rm "$file";
fi
done
rm "{allowed_files}"
rm "{sorted_allowed_files}"
'''])
    # sys.exit()


def clean_all(config):
    for artist in config.artists:
        file = config.adb_push_dir / fs.artist_dir(artist.name).relative_to(fs.music_dir())
        print(f"Cleaning {file}")
        clean(config, artist.name, file)

    for extra_folder in config.extra_sync_folders:
        file = config.adb_push_dir / extra_folder
        print(f"Cleaning {file}")
        clean(config, extra_folder, file)



import shutil

import auto_dlp.downloader as downloader
import auto_dlp.file_cleaning as clean
import auto_dlp.file_locations as fs
import auto_dlp.metadata as metadata
from auto_dlp import playlist_items, terminal_formatting
from auto_dlp.name_cleaning import clean_name


def _copy_if_necessary(src, dest):
    if dest.exists(): return False
    if not src.exists():
        raise RuntimeError(f"There was an unexpected error during download. The file {src} could not be found.")

    fs.touch_file(dest)
    shutil.copyfile(src, dest)
    return True


def download_song(config, artist, name, song_id):
    raw_file = downloader.get_song_file(song_id, config)
    file = fs.best_song_file(raw_file, artist, name)

    fs.touch_folder(file.parent)
    clean.know_file(file.parent)
    necessary = _copy_if_necessary(raw_file, file)
    clean.know_file(file)

    if necessary:
        metadata.write_for_song(config, artist, name, song_id, file)


def download_playlist(config, artist, playlist_name, playlist_id):
    raw_items = playlist_items.get(config, playlist_id)
    items = [
        {"name": clean_name(config, i["name"]), "id": i["id"]} for i in raw_items
    ]

    fs.touch_folder(fs.playlist_dir(artist, playlist_name))
    clean.know_file(fs.playlist_dir(artist, playlist_name))

    for index, name, song_id in fs.iter_playlist_items(artist, playlist_name, items):
        raw_file = downloader.get_song_file(song_id, config)
        if raw_file is None:
            print(
                f"{terminal_formatting.add_color(3, 'Skipping')} "
                f"unavailable song {terminal_formatting.add_color(3, name)} from {playlist_name}")
            continue

        target_path = fs.best_song_file(raw_file, artist, name, playlist_name)

        necessary = _copy_if_necessary(raw_file, target_path)
        clean.know_file(target_path)

        if necessary:
            print(f"Downloaded {name} by {artist} from {playlist_name}")
            metadata.write_for_playlist_item(config, artist, playlist_name, name, playlist_id, song_id, index,
                                             len(items), target_path)

    metadata.process_playlist(config, artist, playlist_name, playlist_id, fs.playlist_dir(artist, playlist_name))



from os.path import getsize

import auto_dlp.YoutubeDataAPIv3 as api
import auto_dlp.file_locations as fs


def _cache_file(song_id):
    return fs.channel_cache_songs() / f"{song_id}.txt"


def _get_content(file):
    with open(file) as fhandle:
        content = fhandle.read()
    return content


def get_for_song(song_id):
    file = _cache_file(song_id)
    if file.exists() and getsize(file) > 0:
        return _get_content(file)

    channel = api.get_song_channel(song_id)
    fs.touch_file(file)
    with open(file, "w") as fhandle:
        fhandle.write(channel)

    return channel



import re
import sys
from functools import cache

_illegal_chars = "\\/!*&\":?#|.,"
_n_whitespace_re = re.compile("\\s{2,}")
_indent = " " * 10


@cache
def regex_for_rule(config, rule):
    for value, repl in config.rule_macros.items():
        rule = rule.replace(value, repl)
    return re.compile(rule, flags=re.IGNORECASE)


def get_illegal_fs_chars():
    return _illegal_chars


def name_contains_illegal_chars(name: str):
    return any(((c in name) for c in _illegal_chars))


def clean_name(config, name: str, verbose=False):
    if verbose:
        print(f"Name at start:\n{_indent}{name}\n", file=sys.stderr)

    if name in config.rename:
        if verbose:
            print(f"Name has been force renamed: {name} -> {config.rename[name]}", file=sys.stderr)
        return config.rename[name]

    og_name = name
    name = "".join(filter(lambda c: c not in _illegal_chars, name))
    name = name.replace(" ", " ")

    if verbose:
        print(f"Name after removing illegal chars:\n{_indent}{name}\n", file=sys.stderr)

    for rule in config.name_rules:
        rule: re.Pattern = regex_for_rule(config, rule)
        name = rule.sub("", name)
        if verbose:
            print(f"Name after regex rule {rule.pattern}:\n{_indent}{name}\n", file=sys.stderr)

    name = _n_whitespace_re.sub(" ", name)
    name = name.strip()

    if name == "":
        if not verbose:
            clean_name(config, og_name, verbose=True)
        raise ValueError(f"Name is empty after cleaning: {og_name}")

    return name



from pathlib import Path

import auto_dlp.codecs as codecs
import auto_dlp.variables as var


def _config_file(name):
    def function():
        file_path = Path(var.CONFIG_DIR).expanduser() / name()
        return file_path

    return function


def touch_folder(folder: Path):
    folder.mkdir(parents=True, exist_ok=True)


def touch_file(file: Path):
    if file.exists(): return
    touch_folder(file.parent)
    file.open("x").close()


@_config_file
def config_dir():
    return ""


@_config_file
def yt_api_key_file():
    return "YT_API_KEY.txt"


@_config_file
def download_cache():
    return "download cache"


@_config_file
def download_cache_trash():
    return "download cache (trash)"


@_config_file
def ytdlp_download_cache():
    return "yt-dlp download cache"


@_config_file
def thumbnail_dir():
    return "thumbnail cache"


@_config_file
def adb_installation():
    return "adb"


@_config_file
def ffmpeg_installation():
    return "ffmpeg"


@_config_file
def playlist_item_cache():
    return "playlist item cache"


@_config_file
def channel_cache():
    return "channel cache"


def channel_cache_songs():
    return channel_cache() / "songs"


@_config_file
def unavailable_items_cache():
    return "unavailable items cache"


@_config_file
def adb_push_files():
    return "adb push files.txt"


def thumbnail_file(prefix, obj_id):
    return thumbnail_dir() / prefix / f"{obj_id}.jpg"


def music_dir():
    return Path(".").expanduser()


def artist_dir(artist):
    return music_dir() / artist


def artist_songs_dir(artist):
    return artist_dir(artist) / "songs"


def try_get_song_file(artist, song, playlist: str = None):
    if playlist is None:
        directory = artist_songs_dir(artist)
    else:
        directory = playlist_dir(artist, playlist)

    for fmt in codecs.available_formats():
        path = directory / f"{song}.{fmt}"
        if path.exists():
            return path

    return None


# Adapts to the format of the original
def best_song_file(original: Path, artist, song, playlist: str = None):
    # Repeated code:
    if playlist is None:
        directory = artist_songs_dir(artist)
    else:
        directory = playlist_dir(artist, playlist)

    fmt = original.suffix
    if fmt == ".m4a":
        fmt = ".aac"
    return directory / f"{song}{fmt}"


def playlist_dir(artist, playlist):
    return artist_dir(artist) / playlist


def iter_playlist_items(artist, playlist, playlist_items):
    index = 0

    for entry in playlist_items:
        song_name = entry["name"]
        song_id = entry["id"]

        yield index, song_name, song_id
        index += 1


def try_get_playlist_files(artist, playlist, playlist_items):
    index = 0

    for entry in playlist_items:
        song_name = entry["name"]
        song_id = entry["id"]

        yield index, song_name, song_id, try_get_song_file(artist, song_name, playlist)
        index += 1



import json
import os
from pathlib import Path


def generate_file(folder: Path, name):
    file = folder / name
    if file.exists():
        print(
            f"A {name} file already exists in the directory {folder}, won't overwrite it. Please delete the file manually if you want to regenerate the sample file.")
        return

    file.touch(exist_ok=False)

    match os.name:
        case "nt":
            config_dir = "C:\\ProgramData"  # This does not work sPath("C:") / "ProgramData"
        case "posix":
            config_dir = Path.home() / ".config"
        case _:
            print(f"Unknown OS name: {os.name} so no custom configuration of config path possible, using POSIX default")
            config_dir = Path.home() / ".config"

    config_dir = config_dir / "auto-dlp"

    json_obj = {
        "config dir": str(config_dir),
        "artists": {
            "Disney": {
                "songs": {
                    "Let it go": "L0MK7qz13bU"
                }
            },
            "Billie Eilish": {
                "playlists": {
                    "Dont smile at me": "OLAK5uy_mYmH6NFnJ6IaH3Ln3wGeVimrFk8FXKzcA",
                    "Happier Than Ever": "OLAK5uy_lkHngwJMHlqHsz7ckye6lwrhYmyFTMvM4"
                }
            },
            "Rick Astley": {
                "songs": {
                    "Never gonna give you up": "dQw4w9WgXcQ"
                }
            }
        }
    }

    with open(file, "w") as fhandle:
        json.dump(json_obj, fhandle, indent="\t")

    print(f"Successfully created file at {file.resolve()}")



import re

import auto_dlp.downloader as downloader
import auto_dlp.file_locations as fs
from auto_dlp import playlist_items
from auto_dlp.name_cleaning import clean_name


class PlaylistWrapper:
    def __init__(self, config, artist, name, oid):
        self.config = config
        self.artist = artist
        self.name = name
        self.oid = oid

    def __str__(self):
        return f"playlist {self.name} ({self.oid}) by {self.artist.name}"

    def items(self):
        for entry in playlist_items.get(config, playlist_id):
            song_name = clean_name(config, entry["name"])
            song_id = entry["id"]

            yield SongWrapper(self.config, self.artist, song_name, song_id, self)

    def delete(self):
        print(f"Deleting {self}:")

        for song in self.items():
            song.delete()


class SongWrapper:
    def __init__(self, config, artist, name, oid, playlist: PlaylistWrapper = None):
        self.config = config
        self.artist = artist
        self.name = name
        self.oid = oid
        self.playlist = playlist

    def __str__(self):
        return f"song {self.name} ({self.oid}) by {self.artist.name}"

    def get_file(self):
        if self.playlist is None:
            return fs.try_get_song_file(self.artist.name, self.name)
        else:
            return fs.try_get_song_file(self.artist.name, self.name, self.playlist.name)

    def delete(self):
        file = self.get_file()
        if file is None: return
        print(f"Deleting {self}: {file}")
        return
        file.unlink(missing_ok=True)
        downloader.delete_cached_version(self.oid)


# Yields: song name, song id and artist
def _iter_songs(config, artist=None):
    if artist is None:
        for artist in config.artists:
            yield from _iter_songs(config, artist)

    for name, song_id in artist.songs.items():
        yield SongWrapper(config, artist, name, song_id)

    for playlist in _iter_playlists(config, artist):
        for entry in playlist_items.get(config, playlist.oid):
            name = clean_name(config, entry["name"])
            song_id = entry["id"]

            yield SongWrapper(config, artist, name, song_id, playlist)


def _iter_playlists(config, artist=None):
    if artist is None:
        for artist in config.artists:
            yield from _iter_playlists(config, artist)

    for playlist_name, playlist_id in artist.playlists.items():
        yield PlaylistWrapper(config, artist, playlist_name, playlist_id)


def _get_obj_locations(config, obj_id):
    obj_re = re.compile(obj_id, flags=re.IGNORECASE)

    for song in _iter_songs(config):
        if song.oid == obj_id or obj_re.fullmatch(song.name) is not None:
            yield song

    for playlist in _iter_playlists(config):
        if playlist.oid == obj_id or obj_re.fullmatch(playlist.name) is not None:
            yield playlist


def redownload(config, obj):
    matches = list(_get_obj_locations(config, obj))
    print(f"Found the following matches for {obj}:")
    for obj in matches:
        print(f"Found {obj}")

    confirmation = input(f"Really redownload all matches for {obj}? (y/n) ")
    if confirmation.strip() != "y":
        print("Aborted")
        return

    for obj in matches:
        obj.delete()



def lazy(initializer):
    value = [None]

    def function():
        if value[0] is None:
            value[0] = initializer()
            if value[0] is None:
                raise ValueError("Lazy property initializer returned None")
        return value[0]

    return function


# The most secure form of encryption know to humankind
def scramble(string: str, encoding="ascii"):
    binary = string.encode(encoding=encoding)

    binary = bytes(reversed(tuple(
        byte ^ 0b0001_1011
        for byte in binary
    )))

    return binary.decode(encoding=encoding)


def take(count, iterable):
    iterable = iter(iterable)
    lst = []

    for i in range(count):
        try:
            lst.append(next(iterable))
        except StopIteration:
            break

    return lst


def print_temp(string):
    hide_temp()
    print(string, end="", flush=True)


def hide_temp():
    print("\x1b[1G\x1b[2K", end="", flush=True)



import os
import sys
import zipfile
from pathlib import Path
from platform import system
from urllib import request

import auto_dlp.file_locations as fs
from auto_dlp import utils

match system():
    case "Linux":
        EXECUTABLE_NAME = None
        DOWNLOAD_URL = None
    case "Windows":
        EXECUTABLE_NAME = Path("ffmpeg-master-latest-win64-gpl-shared") / "bin" / "ffmpeg.exe"
        DOWNLOAD_URL = "https://github.com/BtbN/FFmpeg-Builds/releases/download/latest/ffmpeg-master-latest-win64-gpl-shared.zip"
    case _:
        print(f"This program has not been configured to run on {system()} yet")
        sys.exit(-1)


def check_ffmpeg_availability_linux():
    if os.system("ffmpeg -version") != 0:
        print(
            "Something is wrong with your ffmpeg installation and you are on Linux\nYou are on your own (try sudo apt install ffmpeg)",
            file=sys.stderr)
        sys.exit(-1)


@utils.lazy
def EXECUTABLE():
    if system() == "Linux":
        check_ffmpeg_availability_linux()
        return ""

    installation_path: Path = fs.ffmpeg_installation()
    path = installation_path / EXECUTABLE_NAME

    if path.exists():
        print(f"ffmpeg found at {path}")
        return path

    print(f"Installing ffmpeg for {system()}")

    fs.touch_folder(installation_path)

    (zip_path, _) = request.urlretrieve(DOWNLOAD_URL)

    with zipfile.ZipFile(zip_path, "r") as zf:
        zf.extractall(installation_path)

    return path



from pathlib import PurePosixPath
from subprocess import run

from auto_dlp import terminal_formatting
from auto_dlp.adb import ADB_EXECUTABLE


def _run_command(*command, capture_output=False):
    process = run([ADB_EXECUTABLE(), "-d", "shell", *command], capture_output=capture_output)

    if capture_output:
        return process.stdout


def is_dir(path: PurePosixPath):
    files = list_files(path)
    return len(files) != 1 or files[0] != path


def list_files(path: PurePosixPath):
    output = _run_command("ls", "-1", '"' + str(path.absolute()) + '"', capture_output=True)
    b_file_names = output.split(b"\n")
    file_names = []
    for name in b_file_names:
        try:
            name = name.decode().strip()
        except UnicodeDecodeError as e:
            print(
                f"{terminal_formatting.add_color(3, "Skipping")} file {name} because it could not be decoded using unicode")
            continue

        if name != "":
            file_names.append(name)

    return [path / name for name in file_names]


def delete_file(path: PurePosixPath):
    _run_command("rm", "-r", '"' + str(path.absolute()) + '"')



import argparse
import json
import os
import re
import sys
from pathlib import Path
from traceback import print_exception

import minute_count

import auto_dlp.version
from auto_dlp import config_intepreter, example_file
from auto_dlp.terminal_formatting import add_color

_maybe_wait_after_finishing_execution = True
_definitely_wait_after_finishing_execution = False
CONFIG_FILENAME = "auto-dlp.json"


def manage_music(path: Path, redownload_objects=(), test_names=(), playlist_test_names=None, duration_overview=False,
                 verbose=False):
    if duration_overview:
        minute_count.minute_overview(
            path,
            True,
            True,
            False,
            False,
            True,
            False
        )
        return

    original_cur_dir = Path(os.curdir).resolve()
    os.chdir(path)

    try:
        config_file = path / CONFIG_FILENAME
        if not config_file.exists():
            print(f"No {CONFIG_FILENAME} file found in directory {path.resolve()}", file=sys.stderr)
            print(f"Use auto-dlp DIRECTORY --create-example to create an example {CONFIG_FILENAME} file")
            return

        with open(config_file, encoding="utf-8") as file:
            string = re.sub("//.*", "", file.read())
            json_data = json.loads(string)

        config_intepreter.execute(json_data, redownload_objects=redownload_objects, test_names=test_names,
                                  playlist_test_names=playlist_test_names, verbose=verbose)
    except KeyboardInterrupt:
        print("Program interrupted by user", file=sys.stderr)

    os.chdir(original_cur_dir)


def command_entry_point():
    try:
        main()
    except Exception as e:
        print_exception(e)
        if _maybe_wait_after_finishing_execution:
            wait_to_close()
    except SystemExit:
        if _definitely_wait_after_finishing_execution:
            wait_to_close()


class CustomHelpFormatter(argparse.HelpFormatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, max_help_position=70, width=100)


def main():
    parser = argparse.ArgumentParser(prog="auto-dlp",
                                     description="A light-weight program for managing (and downloading) music",
                                     allow_abbrev=True, add_help=True, exit_on_error=True,
                                     formatter_class=CustomHelpFormatter)
    parser.add_argument("FOLDER",
                        help=f"The folder in which to execute this program. It must contain an {CONFIG_FILENAME} file.")
    parser.add_argument("-rd", "--redownload", metavar="SONG",
                        help="Will delete the current version of the specified song (allows regex), meaning that it will be redownloaded on the next execution of this program")
    parser.add_argument("-rdp", "--redownload-playlist", metavar="PLAYLIST",
                        help="Will delete the current version of the specified playlist (allows regex)")
    parser.add_argument("-tn", "--test-name", metavar="NAME",
                        help="Allows to try out what the renaming system does to a specific string")
    parser.add_argument("-ptn", "--playlist-test-names", metavar="PLAYLIST",
                        help="Runs all the names of songs in PLAYLIST against -tn")
    parser.add_argument("-o", "--overview", action="store_true",
                        help="Generates an overview over the cumulative duration for each folder in the music directory")
    parser.add_argument("-w", "--wait", action="store_true", help="After finishing execution, wait for any input")
    parser.add_argument("-v", "--verbose", action="store_true", help="Makes the script print more information")
    parser.add_argument("--create-example", action="store_true",
                        help=f"Creates a sample {CONFIG_FILENAME} file that will perform elementary functions")
    parser.add_argument("--version", action="store_true", help="Show the version number and exit")

    args = parser.parse_args()
    global _maybe_wait_after_finishing_execution, _definitely_wait_after_finishing_execution
    _maybe_wait_after_finishing_execution = args.wait
    _definitely_wait_after_finishing_execution = args.wait

    if args.version:
        print(f"Auto-DLP version {version.program_version}")
        if args.wait:
            wait_to_close()
        return

    exec_dir = Path(args.FOLDER).expanduser().resolve()
    print(f"Executing Auto-DLP in directory {exec_dir}")

    if args.create_example:
        example_file.generate_file(exec_dir, CONFIG_FILENAME)
        if args.wait:
            wait_to_close()
        return

    redownload = () if args.redownload is None else [args.redownload]
    test_names = () if args.test_name is None else [args.test_name]
    playlist = args.playlist_test_names

    manage_music(exec_dir, redownload_objects=redownload, test_names=test_names, playlist_test_names=playlist,
                 duration_overview=args.overview, verbose=args.verbose)

    if args.wait:
        wait_to_close()


def wait_to_close():
    try:
        input(f"Press {add_color(1, "ENTER")} to close program")
    except KeyboardInterrupt:
        pass



import json
from collections import Counter
from os.path import getsize
from pathlib import Path

import auto_dlp.YoutubeDataAPIv3 as api
import auto_dlp.file_locations as fs
from auto_dlp.name_cleaning import clean_name


def _get_content(file):
    with open(file) as fhandle:
        content = json.load(fhandle)
    return content


# Do not cache; returned objects are mutable
def _get_without_cleaning(config, playlist_id):
    file: Path = fs.playlist_item_cache() / f"{playlist_id}.json"
    if file.exists() and getsize(file) > 0:
        return _get_content(file)

    items = api.get_playlist_items(config, playlist_id)
    fs.touch_file(file)
    with open(file, "w") as fhandle:
        json.dump(items, fhandle)

    return items


# This method operates after name cleaning (unlike the one from YoutubeDataAPIv3)
def deduplicate_playlist_items(config, playlist_items):
    name_counter = Counter()

    # Count occurrences of each name
    for item in playlist_items:
        name_counter[clean_name(config, item["name"])] += 1

    # List of names with count greater one
    index_names = set()

    for name, count in name_counter.items():
        if count > 1:
            index_names.add(name)

    # If everything is fine
    if len(index_names) == 0:
        return playlist_items

    name_counter = Counter()

    for item in playlist_items:
        name = item["name"]
        clean = clean_name(config, name)
        if clean in index_names:
            name_counter[clean] += 1
            item["name"] = f"{name} ={name_counter[clean]}="

    return playlist_items


# Do not cache; returned objects are mutable
def get(config, playlist_id):
    items = _get_without_cleaning(config, playlist_id)

    deduplicate_playlist_items(config, items)

    return items



import shutil
from os.path import getsize
from pathlib import Path

import auto_dlp.file_locations as fs
from auto_dlp import terminal_formatting

_known_files = set()


def know_file(path: Path):
    _known_files.add(path.resolve(strict=True))


def _delete_file(path: Path, reason):
    print(f"Deleting {terminal_formatting.add_color(1, path)} because: {reason}")

    if path.is_dir():
        shutil.rmtree(path)
    else:
        path.unlink()


# def is_known(path: Path, strict_resolve=True):
#     return path.resolve(strict=strict_resolve) in _known_files

def clean(file: Path):
    if file.resolve(strict=True) not in _known_files:
        _delete_file(file, reason="The file was not created by Auto-DLP")
        return

    if not file.is_dir():
        if getsize(file) == 0:
            _delete_file(file, reason="The file is empty")
        return

    for sub in file.iterdir():
        clean(sub)


def clean_all(config):
    for artist in config.artists:
        clean(fs.artist_dir(artist.name))



import os
import re
import sys
from pathlib import Path, PurePosixPath
from time import sleep

import requests

import auto_dlp.adb_fs_cleaning as adb_clean
import auto_dlp.download_manager as download
import auto_dlp.file_cleaning as clean
import auto_dlp.file_locations as fs
import auto_dlp.fs_sync as fs_sync
import auto_dlp.variables as variables
from auto_dlp import adb, playlist_items, age_hints, terminal_formatting
from auto_dlp.file_locations import touch_folder
from auto_dlp.metadata import process_artist
from auto_dlp.name_cleaning import clean_name, name_contains_illegal_chars, get_illegal_fs_chars
from auto_dlp.terminal_formatting import add_color
from auto_dlp.redownloading import redownload


class Config:
    properties = ["config dir", "adb push", "adb push dir", "restart adb", "extra sync folders", "clean push dir",
                  "name rules", "artists", "rule macros", "use cookies", "rename", "name samples", "retry interval",
                  "fragile playlists", "sync dirs", "codec"]

    def __init__(self):
        self.config_dir = None
        self.adb_push = False
        self.adb_push_dir = PurePosixPath("/sdcard/Music/")
        self.restart_adb = False
        self.extra_sync_folders = []
        self.clean_push_dir = True
        self.name_rules = []
        self.compiled_name_rules = []
        self.rule_macros = {}
        self.name_samples = 4
        self.artists = []
        self.use_cookies = False
        self.rename = {}
        self.retry_interval = 10
        self.fragile_playlists = []
        self.sync_dirs = []
        self.invalid_item_names = 0
        self.invalid_item_name_artist_count = 0
        self.codec = "aac"

    def assign_property(self, key, value):
        match key:
            case "config dir":
                self.config_dir = Path(value)
                variables.CONFIG_DIR = self.config_dir
            case "adb push":
                self.adb_push = bool(value)
            case "adb push dir":
                self.adb_push_dir = PurePosixPath(value)
            case "clean push dir":
                self.clean_push_dir = bool(value)
            case "restart adb":
                self.restart_adb = bool(value)
            case "name rules":
                self.name_rules = list(value)
            case "rule macros":
                self.rule_macros = dict(value)
            case "name_samples":
                self.name_samples = int(value)
            case "artists":
                self.artists = [
                    Artist.from_json(self, name, json) for name, json in value.items()
                ]
            case "use cookies":
                self.use_cookies = bool(value)
            case "rename":
                self.rename = dict(value)
            case "retry interval":
                self.retry_interval = float(value)
            case "fragile playlists":
                self.fragile_playlists = list(value)
            case "extra sync folders":
                self.extra_sync_folders = list(value)
            case "sync dirs":
                self.sync_dirs = [Path(path_string).expanduser() for path_string in value]
            case "codec":
                self.codec = str(value)

    def get_locations_to_sync(self):
        for artist in self.artists:
            yield fs.artist_dir(artist.name).resolve()
        for folder in self.extra_sync_folders:
            yield (fs.music_dir() / folder).resolve()

    @classmethod
    def from_json(cls, json):
        config = Config()

        for key, value in json.items():
            if key not in cls.properties:
                raise ValueError(f"Unknown property key: {key}")

            config.assign_property(key, value)

        if config.invalid_item_names > 0:
            print(
                f"{config.invalid_item_names} name(s) over {config.invalid_item_name_artist_count} artist(s) contain an illegal character",
                file=sys.stderr)
            sys.exit(-1)

        return config

    def __str__(self):
        return (f"""config file: {self.config_dir}
adb push: {self.adb_push}
name rules: {" ".join(self.name_rules)}
artists:
""" + "\n".join(map(str, self.artists)))


class Artist:
    properties = ["songs", "playlists"]

    def __init__(self, config, name):
        self.config = config
        self.name = name
        self.songs = {}
        self.playlists = {}

    def all_songs(self):
        yield from self.songs.values()

        for name, playlist_id in self.playlists.items():
            for entry in playlist_items.get(self.config, playlist_id):
                yield entry["id"]

    def assign_property(self, key, value):
        match key:
            case "songs":
                # self.songs = dict(value)
                self.songs = _process_years_in_names(dict(value), age_hints.add_hint_song)
            case "playlists":
                # self.playlists = dict(value)
                self.playlists = _process_years_in_names(dict(value), age_hints.add_hint_playlist)

    def validate_item_names(self):
        invalid_name_count = [0]

        def validate_item_name(item_type: str, name: str):
            if name_contains_illegal_chars(name):
                print(
                    f"{item_type} name contains illegal character(s): {add_color(1, name)} contains one of {add_color(1, get_illegal_fs_chars())}",
                    file=sys.stderr)
                invalid_name_count[0] += 1

        validate_item_name("artist", self.name)
        for song in self.songs.keys():
            validate_item_name("song", song)
        for playlist in self.playlists.keys():
            validate_item_name("playlist", playlist)

        if invalid_name_count[0] > 0:
            self.config.invalid_item_names += invalid_name_count[0]
            self.config.invalid_item_name_artist_count += 1

    @classmethod
    def from_json(cls, config, name, json):
        artist = Artist(config, name)

        for key, value in json.items():
            if key not in cls.properties:
                raise ValueError(f"Unknown property key: {key}")

            artist.assign_property(key, value)

        artist.validate_item_names()

        return artist

    def __str__(self):
        return "\n".join((
            f"""    {self.name}:""",
            *(f"        {name}: {song_id}" for name, song_id in self.songs.items()),
            *(f"        {name}: {playlist_id}" for name, playlist_id in self.playlists.items())
        ))


def _process_years_in_names(mp, hint_function):
    new_map = {}

    for name, obj_id in mp.items():
        new_name = age_hints.parse_name(name, obj_id, hint_function)
        new_map[new_name] = obj_id

    return new_map


def _execute_artist(config, artist):
    print(f"Looking at {terminal_formatting.add_color(2, artist.name)}")
    touch_folder(fs.artist_dir(artist.name))
    clean.know_file(fs.artist_dir(artist.name))

    for song_name, song_id in artist.songs.items():
        print(f"Downloading {song_name} ({song_id})")
        download.download_song(config, artist.name, song_name, song_id)

    process_artist(config, artist, artist.all_songs(), fs.artist_dir(artist.name), fs.artist_songs_dir(artist.name))

    for playlist_name, playlist_id in artist.playlists.items():
        print(
            f"Downloading {playlist_name} ({playlist_id})\n\tinto {fs.playlist_dir(artist.name, playlist_name).resolve()}")
        download.download_playlist(config, artist.name, playlist_name, playlist_id)


def _resiliently_execute_artist(config, artist):
    try:
        _execute_artist(config, artist)
    except requests.exceptions.ConnectionError as e:
        print(f"Connection to Internet services failed: {e}", file=sys.stderr, flush=True)
        print(f"Retrying in {config.retry_interval} seconds")
        sleep(config.retry_interval)
        _resiliently_execute_artist(config, artist)


def execute_adb(config):
    if not config.adb_push: return

    if adb.is_device_connected(config):
        if config.adb_push:
            adb.push_files(config)

        if config.clean_push_dir:
            adb_clean.clean_all(config)
    else:
        print("No Android device connected to this computer using usb", file=sys.stderr)


def execute_sync_sd(config, src: Path, dest: Path):
    print(f"Syncing from {src} to {dest}")
    if not dest.exists():
        print(f"Interrupting sync because destination does not exist: {terminal_formatting.add_color(1, dest)}")
        return
    for to_sync in config.get_locations_to_sync():
        sub_path = to_sync.relative_to(src)
        print(f"Syncing {sub_path}")
        fs_sync.sync(src / sub_path, dest / sub_path)


def execute_sync(config):
    src = Path(os.curdir).resolve()
    for dest in config.sync_dirs:
        execute_sync_sd(config, src, dest)


def execute(json_file, redownload_objects=(), test_names=(), playlist_test_names=None, verbose=False):
    config = Config.from_json(json_file)

    if playlist_test_names is not None:
        for artist in config.artists:
            if playlist_test_names in artist.playlists:
                playlist_test_names = artist.playlists[playlist_test_names]
        test_names = list(test_names)
        test_names += map(lambda x: x["name"], playlist_items.get(config, playlist_test_names))

    if len(test_names) > 0:
        for name in test_names:
            print(f"{name} becomes {clean_name(config, name, verbose=verbose)}")
        return

    if len(redownload_objects) > 0:
        for obj in redownload_objects:
            redownload(config, obj.strip())
        return

    print(config)

    for artist in config.artists:
        _resiliently_execute_artist(config, artist)

    clean.clean_all(config)

    execute_sync(config)
    execute_adb(config)



import auto_dlp.file_locations as fs

_unavailability_err_msgs = ["This video has been removed", "This video is not available", "This video is private",
                            "This video is no longer available", "Video unavailable", "Private video"]


def is_error_message_indicative_of_unavailability(msg):
    return any(i in msg for i in _unavailability_err_msgs)


def _indicator_file(song_id):
    return fs.unavailable_items_cache() / song_id


def know_is_unavailable(song_id):
    if is_unavailable(song_id): return
    fs.touch_file(_indicator_file(song_id))


def is_unavailable(song_id):
    file = _indicator_file(song_id)
    return file.exists()



import stat
import sys
import urllib.request as request
import zipfile
from pathlib import Path
from platform import system
from subprocess import run

import auto_dlp.file_locations as fs
from auto_dlp.utils import lazy

match system():
    case "Linux":
        EXECUTABLE_NAME = Path("platform-tools") / "adb"
        DOWNLOAD_URL = "https://dl.google.com/android/repository/platform-tools-latest-linux.zip"
    case "Windows":
        EXECUTABLE_NAME = Path("platform-tools") / "adb.exe"
        DOWNLOAD_URL = "https://dl.google.com/android/repository/platform-tools-latest-windows.zip"
    case _:
        print(f"This program has not been configured to run on {system()} yet")
        sys.exit(-1)


def _allow_execution(path: Path):
    if system() != "Linux": return

    path.chmod(stat.S_IXUSR)


@lazy
def ADB_EXECUTABLE():
    installation_path: Path = fs.adb_installation()
    path = installation_path / EXECUTABLE_NAME

    if path.exists():
        return path

    print(f"Installing adb for {system()}")

    fs.touch_folder(installation_path)

    (zip_path, _) = request.urlretrieve(DOWNLOAD_URL)

    with zipfile.ZipFile(zip_path, "r") as zip:
        zip.extractall(installation_path)

    _allow_execution(path)

    return str(path)


def validate_names(config):
    illegal_chars = "|\\?*<\":>+[]/"

    def check_name(name):
        if any(c in name for c in illegal_chars):
            print(
                f"The name {name} contains characters ({', '.join(set(illegal_chars) & set(name))}) that can potentially not be pushed to your android device")

    for artist in config.artists:
        check_name(artist.name)
        for playlist in artist.playlists:
            check_name(playlist)


def push_files(config):
    validate_names(config)

    if config.restart_adb:
        print("Restarting adb server")
        run(
            [ADB_EXECUTABLE(), "kill-server"]
        )

    run([
        ADB_EXECUTABLE(), "-d", "push", "--sync",
        # *(fs.artist_dir(a.name).resolve() for a in config.artists),
        # *((fs.music_dir() / folder).resolve() for folder in config.extra_sync_folders),
        *config.get_locations_to_sync(),
        str(config.adb_push_dir)
    ])


def is_device_connected(config):
    if config.restart_adb:
        print("Restarting adb server")
        run(
            [ADB_EXECUTABLE(), "kill-server"]
        )
    return run([ADB_EXECUTABLE(), "shell", "echo", "Connected to android device"]).returncode == 0



import os
import sys
from pathlib import Path

import yt_dlp as yt
from yt_dlp import DownloadError

import auto_dlp.codecs as codecs
import auto_dlp.file_locations as fs
from auto_dlp import unavailable_items, ffmpeg


def _video_url(x):
    return f"https://www.youtube.com/watch?v={x}"


# def _song_file_path(song_id):
#     return fs.download_cache() / f"{song_id}.mp3"

def _try_get_song_file(song_id):
    downld_cache = fs.download_cache()
    for fmt in ["mp3", "m4a"]:
        file = downld_cache / f"{song_id}.{fmt}"
        if file.exists():
            return file
    return None


def _download_song(song_id, config, use_cookies=False):
    if unavailable_items.is_unavailable(song_id):
        return None

    if config.codec not in codecs.available_codecs():
        raise RuntimeError(
            f"In the config the codec to use was specified as {config.codec}, "
            f"but currently only the codecs {codecs.available_codecs()} have been tested and are supported")
    codec = config.codec

    # Saving directory so that current dir can be restored later (has to be changed for yt-dlp to work)
    download_dir = fs.download_cache()
    music_dir = Path(os.curdir).resolve()

    fs.touch_folder(download_dir)
    fs.touch_folder(fs.ytdlp_download_cache())

    os.chdir(download_dir)

    ydl_options = {
        'format': 'm4a/bestaudio/best',
        "cache-dir": fs.ytdlp_download_cache(),
        "outtmpl": "%(id)s",
        "print": f"Song %(id)s has been downloaded",
        'postprocessors': [{
            'key': 'FFmpegExtractAudio',
            'preferredcodec': codec,
        }]
    }

    if ffmpeg.EXECUTABLE() != "":
        ydl_options["ffmpeg_location"] = str(ffmpeg.EXECUTABLE())

    if use_cookies:
        ydl_options["cookiesfrombrowser"] = ("firefox",)

    try:
        with yt.YoutubeDL(ydl_options) as ydl:
            error_code = ydl.download([_video_url(song_id)])
            print(f"Error Code {error_code}")
    except DownloadError as e:
        print("An error occurred during download", file=sys.stderr)

        if unavailable_items.is_error_message_indicative_of_unavailability(e.msg):
            unavailable_items.know_is_unavailable(song_id)
            return None

        if not use_cookies and config.use_cookies:
            print("Retrying with cookies", file=sys.stderr)
            return _download_song(song_id, config, True)
        else:
            return None

    os.chdir(music_dir)

    song_file = _try_get_song_file(song_id)
    if song_file is None:
        raise RuntimeError(
            f"There was an internal error during the download, somehow the file {download_dir / f"{song_id}.{codec}"} was not created")

    return song_file


def _interruptable_download_song(song_id, config):
    try:
        return _download_song(song_id, config)
    except KeyboardInterrupt as e:
        path: Path = _try_get_song_file(song_id)
        if path is not None:
            path.unlink(missing_ok=True)
        raise e


def get_song_file(song_id, config):
    song_file = _try_get_song_file(song_id)
    if song_file is not None:
        return song_file

    music_dir = Path(os.curdir).resolve()
    path = _interruptable_download_song(song_id, config)
    os.chdir(music_dir)
    return path


def delete_cached_version(song_id: str):
    song_file: Path = _try_get_song_file(song_id)
    if song_file is not None:
        # path.unlink(missing_ok=True) # Stuff is no longer just deleted
        fs.download_cache_trash.touch()
        shutil.move(song_file, fs.download_cache_trash / song_file.name)



import shutil
from pathlib import Path

from mutagen.easyid3 import EasyID3

import auto_dlp.file_cleaning as clean
import auto_dlp.thumbnails as thumbnails
from auto_dlp import age_hints
from auto_dlp.terminal_formatting import add_color


def write_for_song(config, artist, name: str, song_id, path: Path):
    mut = EasyID3()
    mut["title"] = name
    mut["artist"] = artist
    mut["albumartist"] = artist
    mut["album"] = f"{artist} songs"
    mut["copyright"] = "none"
    # mut["date"] = "0"
    mut["date"] = str(age_hints.get_hint(None, song_id))
    mut.save(path, v1=2)
    mut.save(path, v1=1)

    # meta = eye.load(path)
    # meta.tag.title = name
    # meta.tag.artist = artist
    # meta.tag.album = f"{artist} songs"
    # meta.tag.album_artist = artist
    # meta.tag.original_release_date = 2000
    #
    # meta.tag.save()
    # meta.tag.track_num = 3


def process_artist(config, artist, songs, artist_path, song_path: Path):
    thumbnail = thumbnails.get_for_artist(config, songs)
    if thumbnail is None: return

    for path in [artist_path, song_path]:
        if not path.exists():
            continue

        img_path = path / "folder.jpg"

        if img_path.exists():
            clean.know_file(img_path)
            continue

        shutil.copyfile(thumbnail, img_path)
        clean.know_file(img_path)


def process_playlist(config, artist, playlist, playlist_id, path: Path):
    img_path = path / "folder.jpg"
    if img_path.exists():
        clean.know_file(img_path)
        return
    try:
        shutil.copyfile(thumbnails.get_for_playlist(playlist_id), img_path)
        clean.know_file(img_path)
    except RuntimeError:
        print(add_color(1, "Could not write playlist thumbnail to playlist directory"))


def write_for_playlist_item(config, artist, playlist, name, playlist_id, song_id, index, item_count, path: Path):
    mut = EasyID3()
    mut["title"] = name
    mut["artist"] = artist
    mut["albumartist"] = artist
    mut["album"] = playlist
    mut["copyright"] = "none"
    mut["tracknumber"] = f"{index + 1}/{item_count}"
    # mut["date"] = "0"
    mut["date"] = str(age_hints.get_hint(playlist_id, song_id))
    mut.save(path, v1=2)
    mut.save(path, v1=1)

    # meta = eye.load(path)
    # meta.tag.title = name
    # meta.tag.artist = artist
    # meta.tag.album = f"{artist} songs"
    # meta.tag.album_artist = artist
    # meta.tag.original_release_date = 2000
    # meta.tag.track_num = index + 1
    #
    # meta.tag.save()



import re
import sys

_name_pattern = re.compile("(?P<Name>.+)\\((?P<Year>[0-9]+)\\)")
_playlist_hints = {}
_song_hints = {}


def add_hint_playlist(playlist_id, year):
    if playlist_id in _playlist_hints:
        print(f"Age hint has already been set for {playlist_id}", file=sys.stderr)
    _playlist_hints[playlist_id] = year


def add_hint_song(song_id, year):
    if song_id in _song_hints:
        print(f"Age hint has already been set for {song_id}", file=sys.stderr)
    _song_hints[song_id] = year


def parse_name(ex_name, obj_id, hint_function):
    match = _name_pattern.fullmatch(ex_name)
    if match is None:
        return ex_name

    values = match.groupdict()

    hint_function(obj_id, int(values["Year"]))

    return values["Name"].strip()


def get_hint(playlist_id, song_id):
    if playlist_id is not None and playlist_id in _playlist_hints:
        return _playlist_hints[playlist_id]

    try:
        return _song_hints[song_id]
    except KeyError:
        return 0



import os.path as path
import re
import sys
from collections import Counter
from json import dumps

from requests import get

import auto_dlp.file_locations as fs
import auto_dlp.utils as utils

# def _ask_for_api_key():
#     key = input("Please enter a valid Google Cloud Youtube API key (This is for getting the playlist item (in the future there might be a different implementation for this feature)): ").strip()
#     if key == "":
#         return _ask_for_api_key()
#
#     fs.touch_file(fs.yt_api_key_file())
#     with fs.yt_api_key_file().open("w") as file:
#         print(key.strip(), file=file)
#
#     return key

# Wow, such secure
_key_url = "https://pastebin.com/8vmQFLbm"


def _download_api_key():
    response = get(_key_url)
    if response.status_code != 200:
        raise RuntimeError("Could not reach server to request api key, please try again later")

    pat = re.compile("this is where the key is:(?P<api_key>.*)not any further", flags=re.IGNORECASE)
    match = pat.search(response.text)
    key = match.groupdict()["api_key"]

    file = fs.yt_api_key_file()
    fs.touch_file(file)
    with open(file, "w") as fhandle:
        fhandle.write(key)

    return key


@utils.lazy
def API_KEY():
    file = fs.yt_api_key_file()
    if not path.exists(file) or path.getsize(file) <= 0:
        print("Api key file was not found", file=sys.stderr, flush=True)
        return utils.scramble(_download_api_key()).strip()

    with open(file) as fhandle:
        return utils.scramble(fhandle.read()).strip()


def request(url, params):
    params["key"] = API_KEY()
    response = get(url, params)

    if response.status_code == 200 and "error" not in response.json():
        return response

    try:
        err_msg = response.json()["error"]["message"]
    except KeyError:
        err_msg = "Error message could not be read"

    if response.status_code == 403 or "API key not valid" in err_msg:
        print("API key was not recognised by Youtube")
        sys.exit()

    raise RuntimeError(f"Network error: {response.status_code} {err_msg}")


# This method is unaware of name cleaning
def deduplicate_playlist_items(playlist_items):
    name_counter = Counter()

    for item in playlist_items:
        name_counter[item["name"]] += 1

    index_names = set()

    for name, count in name_counter.items():
        if count > 1:
            index_names.add(name)

    if len(index_names) == 0:
        return playlist_items

    name_counter = Counter()

    for item in playlist_items:
        name = item["name"]
        if item["name"] in index_names:
            name_counter[name] += 1
            item["name"] = f"{name} ={name_counter[name]}="

    return playlist_items


# currently unused
# def error_on_name_conflict(config, names, playlist_id):
#     clean_names = defaultdict(lambda: [])
#
#     for name in names:
#         clean_names[clean_name(config, name)].append(name)
#
#     conflict_found = False
#
#     for clean, sources in clean_names.items():
#         if len(sources) > 1:
#             conflict_found = True
#             print(f"The specified name rules caused a name conflict: All of\n    {"\n    ".join(sources)}\n    became {clean}")
#
#     if conflict_found:
#         print(f"There are name conflicts in playlist {playlist_id}", file=sys.stderr)
#         sys.exit(-1)

def get_playlist_items(config, playlist_id):
    def send_request(params):
        params["playlist_id"] = playlist_id
        params["part"] = "contentDetails, snippet"
        params["maxResults"] = 50
        response = request("https://www.googleapis.com/youtube/v3/playlistItems", params)
        items = {}

        try:
            items_json = response.json()["items"]

            for item in items_json:
                pos = item["snippet"]["position"]
                items[pos] = {
                    "name": item["snippet"]["title"],
                    "id": item["contentDetails"]["videoId"]
                }
        except KeyError:
            print(dumps(response.json(), indent=4), file=sys.stderr)
            raise RuntimeError("Failed getting playlist items")

        if "nextPageToken" in response.json():
            return response.json()["pageInfo"]["totalResults"], response.json()["nextPageToken"], items
        return response.json()["pageInfo"]["totalResults"], None, items

    received_items = 0
    result_count, page_token, items_map = send_request({})

    items_list = [None] * result_count

    def place_items():
        for index, item in items_map.items():
            if items_list[index] is None:
                items_list[index] = item
                continue

            raise RuntimeError("The Youtube Api returned the same result twice")

    place_items()
    received_items += 50

    while received_items < result_count:
        _, page_token, items_map = send_request({"pageToken": page_token})
        try:
            place_items()
        except RuntimeError as e:
            if playlist_id in config.fragile_playlists:
                return deduplicate_playlist_items(list(filter(lambda x: x is not None, items_list)))
            raise e
        received_items += 50

    result = deduplicate_playlist_items(items_list)

    # error_on_name_conflict(config, [item["name"] for item in result], playlist_id)

    return result


def get_song_channel(song_id):
    response = request("https://www.googleapis.com/youtube/v3/videos", {
        "key": API_KEY(),
        "part": "snippet",
        "id": song_id
    })

    try:
        channel = response.json()["items"][0]["snippet"]["channelId"]
    except KeyError:
        print(dumps(response.json(), indent=4), file=sys.stderr)
        raise RuntimeError("Failed getting song thumbnails")

    return channel


def _get_thumbnails(item_name, api_name, obj_id):
    response = request(f"https://www.googleapis.com/youtube/v3/{api_name}", {
        "key": API_KEY(),
        "maxResults": 50,
        "part": "snippet",
        "id": obj_id
    })

    try:
        thumbnails = response.json()["items"][0]["snippet"]["thumbnails"]
    except (KeyError, IndexError):
        print(dumps(response.json(), indent=4), file=sys.stderr)
        raise RuntimeError(f"Failed getting {item_name} thumbnail for ({obj_id})")

    return thumbnails


def get_song_thumbnails(song_id):
    # response = request("https://www.googleapis.com/youtube/v3/videos", {
    #     "key": API_KEY(),
    #     "maxResults": 50,
    #     "part": "snippet",
    #     "id": song_id
    # })
    #
    # try:
    #     thumbnails = response.json()["items"][0]["snippet"]["thumbnails"]
    # except (KeyError, IndexError):
    #     print(dumps(response.json(), indent=4), file=sys.stderr)
    #     raise RuntimeError("Failed getting song thumbnail")
    #
    # return thumbnails
    return _get_thumbnails("song", "videos", song_id)


def get_playlist_thumbnails(playlist_id):
    # response = request("https://www.googleapis.com/youtube/v3/playlists", {
    #     "key": API_KEY(),
    #     "maxResults": 50,
    #     "part": "snippet",
    #     "id": playlist_id
    # })
    #
    # try:
    #     thumbnails = response.json()["items"][0]["snippet"]["thumbnails"]
    # except (KeyError, IndexError):
    #     print(dumps(response.json(), indent=4), file=sys.stderr)
    #     raise RuntimeError("Failed getting playlist thumbnails")
    #
    # return thumbnails
    return _get_thumbnails("playlist", "playlists", playlist_id)


def get_channel_thumbnails(channel_id):
    # response = request("https://www.googleapis.com/youtube/v3/channels", {
    #     "key": API_KEY(),
    #     "maxResults": 50,
    #     "part": "snippet",
    #     "id": channel_id
    # })
    #
    # try:
    #     thumbnails = response.json()["items"][0]["snippet"]["thumbnails"]
    # except (KeyError, IndexError):
    #     print(dumps(response.json(), indent=4), file=sys.stderr)
    #     raise RuntimeError("Failed getting channel thumbnail")
    #
    # return thumbnails
    return _get_thumbnails("artist", "channels", channel_id)



import urllib.request as request
from collections import Counter
from os.path import getsize
from pathlib import Path

import auto_dlp.YoutubeDataAPIv3 as api
import auto_dlp.file_locations as fs
from auto_dlp import utils, channels


def _get_url(api_function, obj_id):
    thumbnails: dict = api_function(obj_id)
    for key in ["high", "medium"]:
        try:
            return thumbnails[key]["url"]
        except KeyError:
            continue
    for key, value in thumbnails.items():
        return value["url"]


def _get_path(api_function, prefix, obj_id):
    path: Path = fs.thumbnail_file(prefix, obj_id)
    if path.exists() and getsize(path) > 0:
        return path

    fs.touch_file(path)

    url = _get_url(api_function, obj_id)

    request.urlretrieve(url, path)

    return path


def _get(api_function, prefix, obj_id):
    return _get_path(api_function, prefix, obj_id)


def get_for_song(song_id):
    return _get(api.get_song_thumbnails, "songs", song_id)


def get_for_playlist(playlist_id):
    return _get(api.get_playlist_thumbnails, "playlists", playlist_id)


def get_for_artist(config, songs: list):
    samples = utils.take(config.name_samples, songs)
    if len(samples) == 0: return None
    channel_counts = Counter()

    for sample in samples:
        channel_counts[channels.get_for_song(sample)] += 1

    (channel, count) = channel_counts.most_common(1)[0]

    return _get(api.get_channel_thumbnails, "channels", channel)
